-- This file is required file for icekit.exe; icekit.exe loads this
-- file upon startup. see aice_util.txt for function descriptions

require("aice");
require("bit");
require("bit32");

function printf(format, ...)
	io.write(string.format(format, ...))

	-- io.output():setvbuf("line") doesn't seem to work reliably.
	if string.match(format, "\n") ~= nil then
		io.flush()
	end
end

function exit(...)
	os.exit(...)
end

function eq(w1, w2)
	return aice.word_eq(w1, w2)
end

function neq(w1, w2)
	return not aice.word_eq(w1, w2)
end

function tohex(...)
	return bit.tohex(...)
end

-- portability provision layer for LUA 5.2
bit_not = bit32.bnot
bit_and = bit32.band
bit_or = bit32.bor
bit_xor = bit32.bxor
bit_rshift = bit32.rshift
bit_arshift = bit32.arshift
bit_lshift = bit32.lshift
byte_reverse = bit.bswap

function def_edmsr_index(major, minor, ext)
	x = (minor % 16) * 8 + (ext%8)
	return x
end
function def_sys_reg_index(major, minor, ext)
	x = (major%8) * 128 + (minor % 16) * 8 + (ext%8)
	return x
end

_edmsr = {}
_misc = {}
_actrl = {}
_sys = {}

for i = 0,7 do
_edmsr["BPC"   .. i] = def_edmsr_index(3, 0, i)
_edmsr["BPA"   .. i] = def_edmsr_index(3, 1, i)
_edmsr["BPAM"  .. i] = def_edmsr_index(3, 2, i)
_edmsr["BPV"   .. i] = def_edmsr_index(3, 3, i)
_edmsr["BPCID" .. i] = def_edmsr_index(3, 4, i)
end

_edmsr["EDM_CFG"] = def_edmsr_index(3, 5, 0)
_edmsr["EDMSW"] = def_edmsr_index(3, 6, 0)
_edmsr["EDM_CTL"] = def_edmsr_index(3, 7, 0)
_edmsr["EDM_DTR"] = def_edmsr_index(3, 8, 0)
_edmsr["BPMTC"] = def_edmsr_index(3, 9, 0)
_edmsr["DIMBR"] = def_edmsr_index(3, 10, 0)
_edmsr["TECR0"] = def_edmsr_index(3, 14, 0)
_edmsr["TECR1"] = def_edmsr_index(3, 14, 1)
_edmsr["DTR"] = def_edmsr_index(3, 8, 0) -- alias

_misc["DIMIR"] = 0
_misc["SBAR"]  = 1
_misc["EDM_CMDR"] = 2
_misc["DBGER"] = 3
_misc["ACC_CTL"] = 4

_sys["CPU_VER"] = def_sys_reg_index(0, 0, 0)
_sys["ICM_CFG"] = def_sys_reg_index(0, 1, 0)
_sys["DCM_CFG"] = def_sys_reg_index(0, 2, 0)
_sys["MMU_CFG"] = def_sys_reg_index(0, 3, 0)
_sys["MSC_CFG"] = def_sys_reg_index(0, 4, 0)
_sys["CORE_ID"] = def_sys_reg_index(0, 0, 1)
_sys["FUCOP_EXIST"] = def_sys_reg_index(0, 5, 0)
_sys["PSW"] = def_sys_reg_index(1, 0, 0)
_sys["IPSW"] = def_sys_reg_index(1, 0, 1)
_sys["P_IPSW"] = def_sys_reg_index(1, 0, 2)
_sys["IVB"] = def_sys_reg_index(1, 1, 1)
_sys["EVA"] = def_sys_reg_index(1, 2, 1)
_sys["P_EVA"] = def_sys_reg_index(1, 2, 2)
_sys["ITYPE"] = def_sys_reg_index(1, 3, 1)
_sys["P_ITYPE"] = def_sys_reg_index(1, 3, 2)
_sys["MERR"] = def_sys_reg_index(1, 4, 1)
_sys["IPC"] = def_sys_reg_index(1, 5, 1)
_sys["P_IPC"] = def_sys_reg_index(1, 5, 2)
_sys["OIPC"] = def_sys_reg_index(1, 5, 3)
_sys["P_P0"] = def_sys_reg_index(1, 6, 2)
_sys["P_P1"] = def_sys_reg_index(1, 7, 2)
_sys["INT_MASK"] = def_sys_reg_index(1, 8, 0)
_sys["INT_PEND"] = def_sys_reg_index(1, 9, 0)
_sys["MMU_CTL"] = def_sys_reg_index(2, 0, 0)
_sys["L1_PPTB"] = def_sys_reg_index(2, 1, 0)
_sys["TLB_VPN"] = def_sys_reg_index(2, 2, 0)
_sys["TLB_DATA"] = def_sys_reg_index(2, 3, 0)
_sys["TLB_MISC"] = def_sys_reg_index(2, 4, 0)
_sys["VLPT_IDX"] = def_sys_reg_index(2, 5, 0)
_sys["ILMB"] = def_sys_reg_index(2, 6, 0)
_sys["DLMB"] = def_sys_reg_index(2, 7, 0)
_sys["CACHE_CTL"] = def_sys_reg_index(2, 8, 0)
_sys["HSMP_SADDR"] = def_sys_reg_index(2, 9, 0)
_sys["HSMP_EADDR"] = def_sys_reg_index(2, 9, 1)
for i = 0,7 do
_sys["BPC"   .. i] = def_sys_reg_index(3, 0, i)
_sys["BPA"   .. i] = def_sys_reg_index(3, 1, i)
_sys["BPAM"  .. i] = def_sys_reg_index(3, 2, i)
_sys["BPV"   .. i] = def_sys_reg_index(3, 3, i)
_sys["BPCID" .. i] = def_sys_reg_index(3, 4, i)
end

_sys["EDM_CFG"] = def_sys_reg_index(3, 5, 0)
_sys["EDMSW"] = def_sys_reg_index(3, 6, 0)
_sys["EDM_CTL"] = def_sys_reg_index(3, 7, 0)
_sys["EDM_DTR"] = def_sys_reg_index(3, 8, 0)
_sys["BPMTC"] = def_sys_reg_index(3, 9, 0)
_sys["DIMBR"] = def_sys_reg_index(3, 10, 0)
_sys["TECR0"] = def_sys_reg_index(3, 14, 0)
_sys["TECR1"] = def_sys_reg_index(3, 14, 1)
_sys["DTR"] = def_sys_reg_index(3, 8, 0) -- alias

_sys["PFMC0"] = def_sys_reg_index(4, 0, 0)
_sys["PFMC1"] = def_sys_reg_index(4, 0, 1)
_sys["PFMC2"] = def_sys_reg_index(4, 0, 2)
_sys["PFM_CTL"] = def_sys_reg_index(4, 1, 0)

_sys["DMA_CFG"] = def_sys_reg_index(5, 0, 0)
_sys["DMA_GCSW"] = def_sys_reg_index(5, 1, 0)
_sys["DMA_CHNSEL"] = def_sys_reg_index(5, 2, 0)
_sys["DMA_ACT"] = def_sys_reg_index(5, 3, 0)
_sys["DMA_SETUP"] = def_sys_reg_index(5, 4, 0)
_sys["DMA_ISADDR"] = def_sys_reg_index(5, 5, 0)
_sys["DMA_ESADDR"] = def_sys_reg_index(5, 6, 0)
_sys["DMA_TCNT"] = def_sys_reg_index(5, 7, 0)
_sys["DMA_STATUS"] = def_sys_reg_index(5, 8, 0)
_sys["DMA_2DSET"] = def_sys_reg_index(5, 9, 0)
_sys["DMA_2DSCTL"] = def_sys_reg_index(5, 9, 1)

_sys["PRUSR_ACC_CTL"] = def_sys_reg_index(4, 4, 0)
_sys["FUCOP_CTL"] = def_sys_reg_index(4, 5, 0)

_sys["SDZ_CTL"] = def_sys_reg_index(2, 15, 0)
_sys["MISC_CTL"] = def_sys_reg_index(2, 15, 1)

-- for read
_actrl["ICE_STATUS"] = 0x0
_actrl["HW_VER"] = 0x1
_actrl["FPGA_VER"] = 0x2
_actrl["FW_VER"] = 0x3
_actrl["PIN_STATUS"] = 0x4

_actrl["TCK_CTRL"] = 0x0
_actrl["JTAG_CTRL"] = 0x1
_actrl["CLEAR_TIMEOUT"] = 0x2

function s(name)
	return _edmsr[string.upper(name)];
end

function m(name)
	return _misc[string.upper(name)];
end

TCK_48MHz = 8
TCK_24MHz = 9
TCK_12MHz = 10
TCK_6MHz  = 11
TCK_3MHz  = 12
TCK_3000KHz = 12 -- alias
TCK_1500KHz = 13
TCK_750KHz = 14
TCK_375KHz = 15

TCK_30MHz   = 0
TCK_15MHz   = 1
TCK_7MHz    = 2
TCK_3750KHz = 3
TCK_1875KHz = 4
TCK_937KHz  = 5
TCK_468KHz  = 6
TCK_234KHz  = 7


function hex_dump(buf)
   for byte=1, #buf, 16 do
      local chunk = buf:sub(byte, byte+15)
      printf('%08X  ',byte-1)
      chunk:gsub('.', function (c) printf('%02X ',string.byte(c)) end)
      io.write(string.rep(' ',3*(16-#chunk)))
      io.write(' ',chunk:gsub('%c','.'),"\n") 
   end
end

function is_aice_mcu()
	local hw_ver
	hw_ver = aice.cmd_read_ctrl(0x1) -- hardware version
	if (aice_cmd_timeout) then
		error("connectivity to AICE broken")
	end
	local aice_mcu = bit_and(hw_ver, 0x10000) ~= 0
	return aice_mcu
end

function select_first_andes_core()
	local ANDES_JTAG_ID = 0x1000063d
	local ANDES_JTAG_ID_MASK = 0xFF000FFF
	local jtag_id_table = {}
	local target
	local v = aice.set_cmd_verbose(0)

	jtag_id_table = aice.cmd_scan_chain();

	if (aice_cmd_timeout) then
		error("[SCAN_CHAIN] can not access the JTAG scan chain\n")
	end

	for i,v in ipairs(jtag_id_table) do
		local id = bit_and(v, ANDES_JTAG_ID_MASK)
		if (eq(id, ANDES_JTAG_ID)) then
			target = i
			printf("JTAG_ID: %s selected\n", tohex(v))
			break
		end
	end

	if (target == nil) then
		error("can not find AndesCore in the JTAG scan chain\n")
	end

	aice.cmd_select_target(target)

	if (aice_cmd_timeout) then
		error("[SELECT_TARGET] can not select AndesCore for debugging\n")
	end

	aice.set_cmd_verbose(v)
	return target
end

function get_aice_version()
	local hw_ver
	local fpga_ver
	local fw_ver
	local aice_mcu
	local aice_mcu_str
	
	local v = aice.set_cmd_verbose(0)
	hw_ver = aice.cmd_read_ctrl(0x1) -- hardware version
	if (aice_cmd_timeout) then
		error("connectivity to AICE broken")
	end
	aice_mcu = bit_and(hw_ver, 0x10000) ~= 0
	if (aice_mcu) then
		aice_mcu_str = "_mcu"
	else
		aice_mcu_str = ""
	end
	hw_ver = bit_and(hw_ver, 0xffff)
	if (aice_cmd_timeout) then
		return 0, 0, 0
	end

	fpga_ver = aice.cmd_read_ctrl(0x2) -- fpga prom version
	fw_ver = aice.cmd_read_ctrl(0x3) -- firmware version

	if (aice_cmd_timeout) then
		error("USB connection or icebox broken?!")
	end

	_G['aice_mcu'] = aice_mcu -- ugly hack to provide global aice_mcu flag
	printf("aice%s version: %d.%d.%d hw_ver:0x%x fw_ver:0x%x fpga_ver:0x%x\n", aice_mcu_str, hw_ver, fw_ver, fpga_ver, hw_ver, fw_ver, fpga_ver)
	aice.set_cmd_verbose(v)
	return hw_ver, fw_ver, fpga_ver, aice_mcu
end

function dbg_set_dbger(value)
	local RETRY_COUNT = 2
	local v = aice.set_cmd_verbose(0)
	local reg, done, retry

	retry = 0;
	done = false
	while not done do
		reg = aice.cmd_write_misc(m('DBGER'), value)
		if (aice_cmd_timeout) then
			retry = retry + 1
			if (retry >= RETRY_COUNT) then
				error("read DBGER timeout");
			end
			aice.cmd_clear_timeout()
		else
			done = true
		end
	end
	aice.set_cmd_verbose(v)
	return reg
end

function dbg_get_dbger()
	local RETRY_COUNT = 2
	local v = aice.set_cmd_verbose(0)
	local reg, done, retry

	retry = 0
	done = false
	while not done do
		reg = aice.cmd_read_misc(m('DBGER'))
		if (aice_cmd_timeout) then
			retry = retry + 1
			if (retry >= RETRY_COUNT) then
				error("read DBGER timeout");
			end
			aice.cmd_clear_timeout()
		else
			done = true
		end
	end
	aice.set_cmd_verbose(v)
	return reg
end

function dbg_get_dtr()
	local v
	local reg
	v = aice.set_cmd_verbose(0)
	reg = aice.cmd_read_dtr()
	aice.set_cmd_verbose(v)

	return reg
end


function dbg_set_dtr(value)
	local v
	v = aice.set_cmd_verbose(0)
	aice.cmd_write_dtr(value)
	aice.set_cmd_verbose(v)
end

function dbg_set_edm_reg(name, value)
	local edmsr_index = s(name)
	local misc_index = m(name)
	local is_dtr
	local v
	v = aice.set_cmd_verbose(0)
	if (name == "EDM_DTR" or name == "DTR") then
		is_dtr = true
	else
		is_dtr = false
	end

	local RETRY_COUNT = 2
	local reg, done, retry

	done = false
	retry = 0
	while not done do
		if (is_dtr) then
			aice.cmd_write_dtr(value)
		elseif (misc_index ~= nil) then
			aice.cmd_write_misc(misc_index, value)
		elseif (edmsr_index ~= nil) then
			aice.cmd_write_edmsr(edmsr_index, value)
		else
			error(name .. " is not a recognized debug register")
		end
		if (aice_cmd_timeout) then
			retry = retry + 1
			if (retry >= RETRY_COUNT) then
				error(string.format("ERROR: TIMEOUT: failed to write %s to %s\n", tohex(value), name));
			end
			aice.cmd_clear_timeout()
		else
			done = true
		end
	end
	aice.set_cmd_verbose(v)
end

function dbg_get_edm_reg(name, value)
	local edmsr_index = s(name)
	local misc_index = m(name)
	local is_dtr;
	local v
	v = aice.set_cmd_verbose(0)

	if (name == "EDM_DTR" or name == "DTR") then
		is_dtr = true
	else
		is_dtr = false
	end

	local RETRY_COUNT = 2
	local reg, done, retry

	retry = 0
	done = false
	while not done do
		if (is_dtr) then
			reg = aice.cmd_read_dtr()
		elseif (misc_index ~= nil) then
			reg = aice.cmd_read_misc(misc_index)
		elseif (edmsr_index ~= nil) then
			reg = aice.cmd_read_edmsr(edmsr_index)
		else
			error(name .. " is not a recognized debug register")
		end
		if (aice_cmd_timeout) then
			retry = retry + 1
			if (retry >= RETRY_COUNT) then
				error(string.format("TIMEOUT reading %s", name))
			end
			aice.cmd_clear_timeout()
		else
			done = true
		end
	end
	aice.set_cmd_verbose(v)
	return reg
end

-- setup DIM. optional argument: dim_base_address. default = 0xfff00000
function dbg_setup_dim(...)
	local dim_base = ...
	local v = aice.set_cmd_verbose(0)
	if (dim_base == nil) then
		dim_base = 0xfff00000
	end

	dbg_set_edm_reg("DIMBR", dim_base)

	local edm_ctl = dbg_get_edm_reg("EDM_CTL")
	edm_ctl = bit_or(edm_ctl, 0x80000000)
	dbg_set_edm_reg("EDM_CTL", edm_ctl)

	dbg_set_dbger(0x2) -- clear DPED
	aice.set_cmd_verbose(v)
end

function dbg_wait_dex(...)
	local RETRY_COUNT = ...
	-- pulling until DBGER.DEX set
	local retry = 0
	if RETRY_COUNT == nil then
		RETRY_COUNT = 100
	end

	local dbger = 0
	while dbger == 0 do
		dbger = dbg_get_dbger()
		dbger = bit_and(dbger, 0x1)
		if (aice_cmd_timeout) then
			error("failed to get DBGER status");
		end
		retry = retry + 1
		if (retry > RETRY_COUNT) then
			return false
		end
		aice.sleep(1)
	end
	return true
end

-- optional argument use_dbgi_pin; when set, use dbgi_n to trigger the debug exception
-- otherwise, use the DBGI_CMD.
-- dbg_break will wait until DEX is set
function dbg_break(...)
	local use_dbgi_pin = ...
	local v = aice.set_cmd_verbose(0)

	if (use_dbgi_pin == nil) then
		use_dbgi_pin = false
	end

	local dbger = dbg_get_dbger()
	dbger = bit_and(dbger, 0x1)
	if dbger ~= 0 then

		aice.set_cmd_verbose(v)
		return
	end
	
	if (use_dbgi_pin) then
		aice.cmd_assert_dbgi()
		if (aice_cmd_timeout) then
			error("TIMEOUT asserting dbgi_n pin")
		end
	else
		dbg_set_edm_reg("EDM_CMDR", 0x0) -- trigger the debug exception
	end

	if not dbg_wait_dex(100) then
		error("failed to trigger debug exception");
	end

	aice.set_cmd_verbose(v)
end

function reverse_endian(table)
	local rt = {}
	for i,v in ipairs(table) do
		rt[i] = byte_reverse(v)
	end
	return rt
end

-- w is an array of 4 words. the words are instructions in big endian format
-- as shown by objdump output.
--
-- CPU must be in the debug state, DIMBR must be setup to use this function
function dbg_execute(w)
	local w = reverse_endian(w)
	
	dbg_set_edm_reg("DIMIR", 0)

	aice.cmd_write_dim(w[1], w[2], w[3], w[4])

	local RETRY_COUNT = 2
	local reg, done, retry
	retry = 0
	done = false
	while not done do
		aice.cmd_execute()
		if (aice_cmd_timeout) then
			retry = retry + 1
			if (retry >= RETRY_COUNT) then
				error("failed to resume execution");
			end
			aice.cmd_clear_timeout()
		else
			done = true
		end
	end

	RETRY_COUNT = 100
	retry = 0
	done = false
	while not done do
		reg = dbg_get_dbger()
		if (aice_cmd_timeout) then
			error("failed to resume execution");
		end
		reg = bit_and(reg, 0x2)
		done = (reg == 0x2)
		retry = retry + 1
		if (retry > 3) then
			aice.sleep(1)
		end
		if (not done and (retry > RETRY_COUNT)) then
			error(string.format("execution not done"))
		end
	end

	-- clear DPEX bit (write one to clear)
	dbg_set_dbger(0x2)
	if (aice_cmd_timeout) then
		error("failed to resume execution");
	end
end

-- This function requires DIM to be setup
-- dbg_get_pc()
--     get PC when the debug exception happened
--
--               1:      ! get IPC
-- 6402A402              mfsr    $r0, $IPC
-- 64070003              mtsr    $r0, $EDM_DTR
-- 64000008              dsb
-- 48fffffa              j 1b
function dbg_get_pc()
	local v = aice.set_cmd_verbose(0)
	local reg
	local r0 = dbg_get_gpr(0)
	dbg_execute({0x6402A402, 0x64070003, 0x64000008, 0x48fffffa})
	reg = dbg_get_dtr()
	if (aice_cmd_timeout) then
		error("failed to get ipc")
	end
	dbg_set_gpr(0, r0)
	aice.set_cmd_verbose(v)
	return reg
end

-- This function requires DIM to be setup
-- dbg_get_pc()
--     get PC when the debug exception happened
--
--               1:      ! set IPC
-- 64070002              mfsr    $r0, $EDM_DTR
-- 6402A403              mtsr    $r0, $IPC
-- 64000008              dsb
-- 48fffffa              j 1b
function dbg_set_pc(value)
	local v = aice.set_cmd_verbose(0)
	local reg
	local r0 = dbg_get_gpr(0)

	dbg_set_dtr(value)
	dbg_execute({0x64070002, 0x6402A403, 0x64000008, 0x48fffffa})
	if (aice_cmd_timeout) then
		error("failed to set ipc")
	end
	dbg_set_gpr(0, r0)
	aice.set_cmd_verbose(v)
	return reg
end

-- This function requires DIM to be setup
-- dbg_get_sys_reg()
--     get system register
--
--               1:      ! get system registers
-- 64000002              mfsr    $r0, $XXX
-- 64070003              mtsr    $r0, $EDM_DTR
-- 64000008              dsb
-- 48fffffa              j 1b
function dbg_get_sys_reg(sys_reg_name)
	local v = aice.set_cmd_verbose(0)
	local reg
	local w = {0x64000002, 0x64070003, 0x64000008, 0x48fffffa}
	local sys_reg_index = _sys[sys_reg_name]
	if sys_reg_index == nil then
		error(string.format("invalid register name: %s\n", sys_reg_name))
	end

	local r0 = dbg_get_gpr(0)
	w[1] = bit_or(w[1], bit_lshift(sys_reg_index, 10))
	dbg_execute(w)
	reg = dbg_get_dtr()
	if (aice_cmd_timeout) then
		error(string.format("failed to get %s", sys_reg_name))
	end
	dbg_set_gpr(0, r0)
	aice.set_cmd_verbose(v)
	return reg
end

-- This function requires DIM to be setup
-- dbg_set_sys_reg()
--     get system register
--
--               1:      ! set system registers
-- 64070002              mfsr    $r0, $EDM_DTR
-- 64000003              mtsr    $r0, $XXX
-- 64000008              dsb
-- 48fffffa              j 1b
function dbg_set_sys_reg(sys_reg_name, value)
	local v = aice.set_cmd_verbose(0)
	local r0 = dbg_get_gpr(0)

	dbg_set_dtr(value)

	local w = {0x64070002, 0x64000003, 0x64000008, 0x48fffffa}
	local sys_reg_index = _sys[sys_reg_name]
	if sys_reg_index == nil then
		error(string.format("invalid register name: %s\n", sys_reg_name))
	end

	w[2] = bit_or(w[2], bit_lshift(sys_reg_index, 10))
	dbg_execute(w)
	if (aice_cmd_timeout) then
		error(string.format("failed to get %s", sys_reg_name))
	end
	dbg_set_gpr(0, r0)
	aice.set_cmd_verbose(v)
	return reg
end

-- dbg_continue()
--     let target resume execution and leaves debug mode
--
-- 40000009              nop
-- 40000009              nop
-- 40000009              nop
-- 64000004              iret
function dbg_continue()
	local v = aice.set_cmd_verbose(0)
	local reg

	dbg_set_dbger(0x1) -- clear DEX

	dbg_execute({0x40000009, 0x40000009, 0x40000009, 0x64000004})
	aice.set_cmd_verbose(v)
end

-- dbg_step()
--     single step one instruction
function dbg_step(...)
	local r0 = ...
	local v = aice.set_cmd_verbose(0)
	local reg

	if r0 == nil then
		r0 = dbg_get_gpr(0) -- backup r0
	end

	local psw = dbg_get_sys_reg('PSW')
	psw = bit_or(psw, 0x800)  -- enable HSS, set DEX
	ipsw = bit_xor(psw, 0x400) -- enable HSS, clear DEX
	dbg_set_sys_reg('IPSW', ipsw)
	dbg_set_sys_reg('PSW', psw)
	dbg_set_gpr(0, r0)	-- restore r0
	dbg_continue()
	if not dbg_wait_dex(200) then
		error("control is not returned to debugger after single stepping\n");
	end

	r0 = dbg_get_gpr(0)
	ipsw = bit_xor(ipsw, 0x800) -- disable HSS
	psw = bit_xor(psw, 0x800) -- disable HSS
	dbg_set_sys_reg('IPSW', ipsw)
	dbg_set_sys_reg('PSW', psw)
	dbg_set_gpr(0, r0)
	aice.set_cmd_verbose(v)
end

function dbg_wait_crst()
	local v = aice.set_cmd_verbose(0)
	local interval = 100
	local dbger = 0
	local retry = 0
	local RETRY_COUNT = 50	-- 50*100 = 5000 ms
	while dbger == 0 do
		dbger = dbg_get_dbger()
		dbger = bit_and(dbger, 0x4)
		aice.sleep(interval)
		retry = retry + 1
		if (retry > RETRY_COUNT) then
			return
		end
	end
	dbg_set_dbger(0x4) -- clear CRST
	aice.set_cmd_verbose(v)
end

function dbg_soft_restart(...) 
	local dim_base, init_sys_rst_delay = ...
	local dummy_cmd = {}
	local v = aice.set_cmd_verbose(0)
	local sysrst_delay = init_sys_rst_delay

	if (dim_base == nil) then
		dim_base = 0xfff00000
	end
	if (sysrst_delay == nil) then
		sysrst_delay = 6
	end

	local RETRY_COUNT = 2
	local done, retry
	retry = 0
	done = false
	while not done do
		aice.cmd_sys_reset()
		-- do not use dbg_wait_crst() here to shorten reset delay
		aice.sleep(sysrst_delay)

		aice.cmd_write_edmsr(s('DIMBR'), dim_base)
		if (aice_cmd_timeout) then
			aice.cmd_clear_timeout()
			if (sysrst_delay < 25) then
				sysrst_delay = sysrst_delay + sysrst_delay / 2 + 1
			elseif (sysrst_delay < 250) then
				sysrst_delay = 2*sysrst_delay + 1
			elseif (sysrst_delay < 1900) then
				sysrst_delay = 2*sysrst_delay + 11
			elseif (sysrst_delay < 5000) then
				sysrst_delay = sysrst_delay + 501
			end
			if (sysrst_delay > 5000) then
				error("can not establish connection with DUT")
			end
		else
			done = true
		end
	end

	RETRY_COUNT = 2
	done = false
	aice.set_usb_verbose(0)
	while not done do
		aice.cmd_write_ctrl(0x01, 0x3) -- sysrst & trst
		if (aice_cmd_timeout) then
			error(string.format("connect to AICE failed"))
		end
		aice.sleep(sysrst_delay)
		aice.cmd_write_edmsr(s('DIMBR'), dim_base)
		if (aice_cmd_timeout) then
			error("failed")
		end
		aice.set_pack_mode(1)
		aice.cmd_write_edmsr(s('EDM_CTL'), 0x80000000)
		aice.cmd_write_misc(m('EDM_CMDR'), 0x0) -- trigger the debug exception
		aice.set_pack_mode(0)

		-- more setup
		aice.cmd_write_misc(m('DIMIR'), 0x0)
		if (aice_cmd_timeout) then
			retry = retry + 1
			if (retry >= RETRY_COUNT) then
				error(string.format("connect to DUT failed"))
				os.exit(1)
			end
			aice.cmd_clear_timeout()
		else
			done = true
		end
	end
	aice.set_usb_verbose(0)
	
	dbg_set_dbger(0x2) -- clear DPED
	aice.set_cmd_verbose(v)
end

--                 1: ! init r1            ! write index
-- 64170002              mfsr $r1, $EDM_DTR
-- 40000009              nop
-- 40000009              nop
-- 48fffffa              j 1b
function dbg_set_gpr(reg_idx, value)
	local v = aice.set_cmd_verbose(0)
	local reg

	dbg_set_dtr(value)

	local w = {0x64070002, 0x40000009, 0x40000009, 0x48fffffa}
	reg_idx = bit_and(reg_idx, 0x1f)
	w[1] = bit_or(w[1], bit_lshift(reg_idx, 20))
	dbg_execute(w)

	aice.set_cmd_verbose(v)
end

--          	1: ! backup r1
-- 64170003              mtsr $r1, $EDM_DTR
-- 64000008              dsb
-- 40000009              nop
-- 48fffffa              j 1b
function dbg_get_gpr(reg_idx)
	local v = aice.set_cmd_verbose(0)
	local reg

	local w = {0x64070003, 0x64000008, 0x40000009, 0x48fffffa}
	reg_idx = bit_and(reg_idx, 0x1f)
	w[1] = bit_or(w[1], bit_lshift(reg_idx, 20))
	dbg_execute(w)

	reg = dbg_get_dtr()
	aice.set_cmd_verbose(v)

	return reg
end

function dbg_load_file(addr, filename)
	local v = aice.set_cmd_verbose(0)
	local image = aice.load_image(filename)
	for i=1, #image do
		local word = image[i]
		local _addr = addr+(i-1)*4
		aice.cmd_write_word(_addr, word)
		if (aice_cmd_timeout) then
			break
		end
	end
	aice.set_cmd_verbose(v)
end

-- #chunk must be <= 256
function dbg_fastwrite_mem(addr, chunk)
	local v = aice.set_cmd_verbose(0)
	-- align to word
	local a = bit_rshift(addr, 2)
	a = bit_lshift(a, 2)
	a = bit_or(a, 1)	-- bit 0 of SBAR indicates this is a fast memory write

	dbg_set_edm_reg('SBAR', a)
	if (aice_cmd_timeout) then
		return
	end
	aice.cmd_fastwrite_mem(chunk)
	aice.set_cmd_verbose(v)
end

function dbg_fast_load_file(addr, filename)
	local image = aice.image(filename)
	local k
	k = 0
	while (k < #image) do
		local chunk = {}
		for i=1,256 do
			k = k + 1
			if (k > #image) then
				break
			end
			chunk[i] = image[k]
		end
		do_fastwrite_mem(addr, chunk)
		addr = addr + (#chunk*4)
		if (aice_cmd_timeout) then
			break
		end
	end
end

function dbg_block_dump(addr, size)
	local v = aice.set_cmd_verbose(0)
	for a = addr, addr+size-1, 4 do
		w = aice.cmd_read_word(a)
		if (aice_cmd_timeout) then
			break
		end
		printf("[%s] = 0x%s\n", tohex(a), tohex(w))
	end
	aice.set_cmd_verbose(v)
end

function opt_clear_timeout()
	local v = aice.set_cmd_verbose(0)
	local reg = 0x10 -- RDY = 1

	if not aice_cmd_timeout then
		reg = aice.cmd_read_ctrl(0x00)	-- only valid for AICE, not AICE_MCU
		if (aice_cmd_timeout) then
			error("USB connection or icebox broken?!\n")
		end

		if (reg == nil) then
			reg = 0
		else
			reg = bit_and(reg, 0x10)
		end
	else
		reg = 0x0
	end

	if (reg == 0) then
		aice.cmd_clear_timeout()
	end

	if (aice_cmd_timeout) then
		error("USB connection or icebox broken?!\n")
	end

	aice.set_cmd_verbose(v)
end

function clear_timeout()
	local v = aice.set_cmd_verbose(0)

	aice.cmd_clear_timeout()
	if (aice_cmd_timeout) then
		error("USB connection or icebox broken?!\n")
	end

	aice.set_cmd_verbose(v)
end

function check_timeout(...)
	if (aice_cmd_timeout) then
		error(...)
	end
end

cmd_read_mem = cmd_read_word
cmd_write_mem = cmd_write_word
